@@ -1,32 +1,42 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright 2019 by Vinay Sajip. All Rights Reserved.
+# Copyright 2019-2022 by Vinay Sajip. All Rights Reserved.
 #
-
+from __future__ import print_function
+import datetime
 import inspect
 import io
+import json
 import logging
 from os import path, remove, close
+import re
 from shutil import rmtree
+import sys
 import tempfile
 import xml.etree.ElementTree as ET
 
 try:
-    from urllib.parse import urljoin
+    from urllib.parse import urljoin, urlencode
+    from urllib.request import urlopen, Request
     basestring = str
 except ImportError:
     from urlparse import urljoin
+    from urllib import urlencode
+    from urllib2 import urlopen, Request
+
+from docutils.nodes import (strong, emphasis, inline, Text, document,
+                            paragraph, reprunicode, raw)
 
-from docutils.nodes import strong, emphasis, inline, Text, document
 from docutils.parsers.rst.roles import set_classes
 
 # from sphinx import version_info as sphinx_version
+from sphinx.addnodes import literal_strong
 from sphinx.util.console import bold
 from sphinx.writers.html import logger, HTMLTranslator as BaseTranslator
 
 HERE = path.abspath(path.dirname(__file__))
 
-__version__ = '0.0.8'
+__version__ = '0.1.0'
 
 logger = logging.getLogger(__name__)
 logger.addHandler(logging.NullHandler())
@@ -57,6 +67,55 @@ def generic_span(role, rawText, text, lineno, inliner, options=None, context=Non
     node.append(content)
     return [node], []
 
+_ICONIFY_PATTERN = re.compile(r'^(?P<icon_set>[\w-]+):(?P<icon>[\w-]+)(,(?P<height>\d+(\.\d+)?(px|r?em)))?(,(?P<classes>#?[\w-]+(,[\w-]+)*))?$')
+
+_REMOVE_PATTERN = re.compile(r'xmlns(:xlink)?="[^"]+" ')
+
+_ICON_DATA = {}
+
+def iconify(role, rawText, text, lineno, inliner, options=None, context=None):
+    if text in _ICON_DATA:
+        svg = _ICON_DATA[text]
+    else:
+        m = _ICONIFY_PATTERN.match(text)
+        if not m:
+            raise ValueError('Invalid icon pattern: %r' % text)
+        d = m.groupdict()
+        url = 'https://api.iconify.design/%s/%s.svg' % (d['icon_set'], d['icon'])
+        params = {}
+        if d['height']:
+            params['height'] = d['height']
+        # import pdb; pdb.set_trace()
+        if params:
+            qs = urlencode(params)
+            url = '%s?%s' % (url, qs)
+        req = Request(url, headers={'User-Agent': 'sphinx-sizzle-theme'})
+        resp = urlopen(req)
+        # resp = requests.get(url, params=params)
+        if resp.status != 200:
+            s = '%s:%s' % (d['icon_set'], d['icon'])
+            raise ValueError('Icon not found: %r' % s)
+        s = resp.read().decode('utf-8')  # Might need to revisit
+        if not d['classes']:
+            attrs = ' class="iconify"'
+        else:
+            parts = d['classes'].split(',')
+            attrs = ''
+            if parts[0].startswith('#'):
+                id = parts.pop(0)
+                attrs = ' id="%s"' % id[1:]
+            if 'iconify' not in parts:
+                parts.append('iconify')
+            attrs += ' class="%s"' % (' '.join(parts))
+        svg = s.replace('<svg', '<svg%s' % attrs)
+        svg = _REMOVE_PATTERN.sub('', svg)
+        _ICON_DATA[text] = svg
+    content = Text(svg)
+    node = raw(format='html')
+    node.append(content)
+    # import pdb; pdb.set_trace()
+    return [node], []
+
 def create_sitemap(app):
     filename = app.outdir + '/sitemap.xml'
     logger.info(bold('creating sitemap... '), nonl=True)
@@ -71,8 +130,34 @@ def create_sitemap(app):
     ET.ElementTree(root).write(filename)
     logger.info('done')
 
+_APP_JS_SNIPPET = '''
+(function() {
+  var sizzle = $(document).data('sizzle');
+
+  sizzle.app_data = JSON.parse('%s');
+})();
+'''
+
+def create_app_data(app):
+    logger.info(bold('creating app data... '), nonl=True)
+    filename = app.outdir + '/app-data.js'
+    data = {
+        'glossary': {
+            'document': app.glossary_doc,
+            'terms': app.glossary_info,
+        },
+        'custom_data': app.config.html_theme_options.get('custom_data', {})
+    }
+    s = json.dumps(data).replace('\\', '\\\\').replace('\'', '\\\'')
+    s = _APP_JS_SNIPPET % s
+    with io.open(filename, 'w', encoding='utf-8') as f:
+        f.write(s)
+    logger.info('done')
+
 def on_init(app):
     app.first_permalinks = {}
+    app.glossary_info = {}
+    app.glossary_doc = None
     if logging_enabled:
         fd, fn = tempfile.mkstemp(prefix='sizzle-', suffix='.log')
         close(fd)
@@ -94,6 +179,8 @@ def on_init(app):
 def on_build_finished(app, exception):
     if app.sitemap_urls:
         create_sitemap(app)
+    create_app_data(app)
+
     # remove unused files.
     outdir = app.builder.outdir
     unused = (
@@ -105,13 +192,13 @@ def on_build_finished(app, exception):
         ('_static', 'comment.png'),
         ('_static', 'comment-bright.png'),
         ('_static', 'comment-close.png'),
-        ('_static', 'file.png'),
+        #('_static', 'file.png'),
         ('_static', 'up.png'),
         ('_static', 'up-pressed.png'),
         ('_static', 'down.png'),
         ('_static', 'down-pressed.png'),
-        ('_static', 'plus.png'),
-        ('_static', 'minus.png'),
+        #('_static', 'plus.png'),
+        #('_static', 'minus.png'),
         ('_static', 'basic.css'),
     )
     for cp in unused:
@@ -138,6 +225,12 @@ def on_page(app, pagename, templatename, context, doctree):
         url = urljoin(base_url, pagename + '.html')
         app.sitemap_urls.append(url)
     context['sizzle_version'] = app.sizzle_version
+    context['build_date'] = datetime.date.today().strftime('%Y-%m-%d')
+    if pagename != 'search':
+        app_data_path = context['pathto']('search').replace('search.html',
+                                                            'app-data.js')
+        if app_data_path != '#':
+            context['app_data_path'] = app_data_path
     logger.debug('on_page: %s', pagename)
 
 def find_first_permalink(document):
@@ -192,6 +285,11 @@ def extract_keys(source, keys):
             result[key] = source[key]
     return result
 
+def dump_node(node, level=0, file=sys.stdout):  # used for debugging only
+    print('%s%r' %('  ' * level, node), file=file)
+    for child in node.children:
+        dump_node(child, level + 1, file=file)
+
 class Translator(BaseTranslator):
     def __init__(self, *args, **kwargs):
         super(Translator, self).__init__(*args, **kwargs)
@@ -208,6 +306,13 @@ class Translator(BaseTranslator):
         self.toctree = toctree
         if toctree:
             logger.debug('translate toctree: %s', self.builder.current_docname)
+        # Remove limit on field names (ensures parameters are in two columns)
+        self.settings.field_name_limit = 0
+        self.in_glossary = self.in_field_body = False
+        ctx = self.builder.globalcontext
+        self.enable_tooltips = ctx.get('theme_enable_tooltips') in (True, 'true')
+        self.glossary_permalinks = ctx.get('theme_glossary_permalinks') in (True, 'true')
+        self.debugging = False
 
     def visit_table(self, node, name=''):
         """
@@ -274,6 +379,8 @@ class Translator(BaseTranslator):
             para.insert(0, handle)
 
     def visit_list_item(self, node):
+        if self.in_field_body:
+            self.fixup_parameter(node[0])
         super(Translator, self).visit_list_item(node)
         self.li_level += 1
 
@@ -317,6 +424,29 @@ class Translator(BaseTranslator):
                 logger.debug('visit_reference: %s -> %s', ru, node['refuri'])
         super(Translator, self).visit_reference(node)
 
+    def fixup_parameter(self, node):
+        # is this a parameter declaration with an oddly-parsed structure such
+        # that we have a literal_strong for the parameter name, an ndash
+        # separator, and then a paragraph? That needs slight massaging so we
+        # don't have a dangling ndash ...
+        if (isinstance(node[0], literal_strong) and
+            (node[1] == ' \u2013 ') and isinstance(node[2], paragraph)):
+            para = node[2]
+            del node[2]
+            #remove the paras children and insert them in place of the para
+            for n in reversed(para.children):
+                node.children.insert(2, n)
+
+    def visit_field_body(self, node):
+        self.in_field_body = True
+        if len(node):
+            self.fixup_parameter(node[0])
+        super(Translator, self).visit_field_body(node)
+
+    def depart_field_body(self, node):
+        super(Translator, self).depart_field_body(node)
+        self.in_field_body = False
+
     def visit_bullet_list(self, node):
         nda = node.non_default_attributes()
         if 'classes' in nda:
@@ -373,6 +503,49 @@ class Translator(BaseTranslator):
                          '<tr>')
         self.footnote_backrefs(node)
 
+    # Glossary processing
+
+    def visit_glossary(self, node):
+        self.in_glossary = True
+        b = self.builder
+        if b.app.glossary_doc is None:
+            b.app.glossary_doc = b.current_docname
+        super(Translator, self).visit_glossary(node)
+
+    def depart_glossary(self, node):
+        super(Translator, self).depart_glossary(node)
+        self.in_glossary = False
+
+    def visit_term(self, node):
+        super(Translator, self).visit_term(node)
+        if self.in_glossary and self.enable_tooltips:
+            self.term_key = node.attributes['ids'][0]
+            self.term_pos = len(self.body)
+
+    def depart_term(self, node):
+        if self.in_glossary and self.enable_tooltips:
+            term_html = ''.join(self.body[self.term_pos:])
+            app = self.builder.app
+            app.glossary_info.setdefault(self.term_key, {})['term'] = term_html
+        if self.in_glossary and self.glossary_permalinks:
+            s = ('<a class="headerlink" href="#%s" '
+                 'title="Permalink to this term">\xb6</a>' % self.term_key)
+            self.body.append(s)
+        super(Translator, self).depart_term(node)
+
+    def visit_definition(self, node):
+        super(Translator, self).visit_definition(node)
+        if self.in_glossary and self.enable_tooltips:
+            self.defn_pos = len(self.body)
+
+    def depart_definition(self, node):
+        if self.in_glossary and self.enable_tooltips:
+            defn_html = ''.join(self.body[self.defn_pos:])
+            app = self.builder.app
+            app.glossary_info.setdefault(self.term_key, {})['defn'] = defn_html
+        super(Translator, self).depart_definition(node)
+
+
 def setup(app):
     app.add_html_theme('sizzle', path.abspath(path.dirname(__file__)))
     app.set_translator('html', Translator)
@@ -382,5 +555,6 @@ def setup(app):
     app.connect('doctree-read', on_doctree_read)
     app.connect('doctree-resolved', on_doctree_resolved)
     app.add_role('fa', font_awesome)
+    app.add_role('icon', iconify)
     app.add_role('span', generic_span)
     app.sitemap_urls = []
